function [] = print_route(w12, b2, w23, b3, HiddenFunc, OutFunc, maze_i, maze_j)
% PRINT_ROUTE : Search a route by using the Q table. 
%           Then, display the route.
%
% Syntax : PRINT_ROUTE(Q_table);
%
% Input  : Q_table (Requred) : N x 4 (The number of states x The number of actions)
%
% Output : 
% 
% See also TABLE, DISP, PRINT_QTABLE.
    
    disp('Route:');
    s = 1;                  % Initial State : S1

    for i=1:9             % Maximum length of the route : 9
        st = zeros(maze_i*maze_j, 1);
        st(s) = 1;
        
        % Input (State Number)
        X = st;
        
        z2 = w12.' * X + b2;                % Weighted sum of the hidden layers z2
        if(HiddenFunc == "sigmoid")
            a2 = 1./(1+exp(-z2));               % Output a2 of the hidden layer (activation function: sigmoid function)
        end
        
        z3 = w23.' * a2 + b3;               % Weighted sum of the output layers z3
        
    if(OutFunc == 'Linear')
        a3 = z3;      % Output of output layer a3 (activation function: lenear function)
    end
        
        Qt = a3;
        max_value = max(Qt);
        for j = 1 : 4
            if max_value == Qt(j)
                I = j ;     % Decide the action in state t.
            end
        end
        
        if s == 9          % S9 is the goal. When the goal is reached, it is marked as goal.
            fprintf('(S%d) Goal!\n', s);
            break
        end

        fprintf('(S%d,%d)', s, I);        % Display the state t.

        % The following equation is used to determine whether a state (position) is at the edge of a wall around the maze, and to move to the next state (position) if it is not at the edge.
        % If it bumps into the surrounding walls, it will display Error.
        
        if I == 1                   % If the action of the agent on the state,  
            if rem(s,3) ~= 0        % (that is not the far right) is one (right),
                s = s+1;            % the agent moves to next state (state s+1).
            else
                disp('Error');
                break
            end
        elseif I == 2               % Action : 2 (up), state is not at the top
            if s > 3
                s = s-3;            % the agent moves to the next state (s-3: 3x3)
            else
                disp('Error');
                break
            end
        elseif I == 3               % Action : 3 (left), state is not on the left side
            if rem(s,3) ~= 1
                s = s-1;            % the agent moves to the next state (s-3: 3x3)
            else
                disp('Error');
                break
            end
        elseif I == 4               % Action : 4 (down), state is not at the bottom
            if s < 7
                s = s+3;            % the agent moves to the next state (s+3: 3x3)
            else
                disp('Error');
                break
            end
        end
    end
    
    if i == 25                      % Maximum length of the route : 25
        disp('Error');              % If the longest route does not lead to the goal, Error
    end
    disp(' ');disp(' ');
end