% How to useF
%   sw_Q_Learning s
%
% Required filesFsw_Q_Learning.m, sw_move_state.m, sw_update_Q.m, print_route.m, print_weight.m
%
% Toolbox : MATLAB 
%
% Tool Version: MATLAB R2021a
% >> [fList,pList] = matlab.codetools.requiredFilesAndProducts('tb_Q_Learning.m');
% >> fList{:}
% >> {pList.Name}'
% Or after running
% >> license('inuse')
%
% see also SW_MOVE_STATE, SW_UPDATE_Q, PRINT_ROUTE, PRINT_WEIGHT.
clear all;
close;
rng(1); % Fixed seed value for random numbers

%% Setting the maximum number of episodes, maximum number of actions, and rewards in learning
% The maximum number of episodes.
    MaxEpisode = 10000; 
% The maximum number of actions in one episode.
    Max_t = 10;   
% RwwardiGoalF+5CDemonF-5COtherF0j
    reward_table = [ ...
        0    0     0 ...
        0    -5     0 ...
        -5    -5     10]';

%% Generating a maze board
% Maze Sizeii ~ jj
    maze_i = 3;
    maze_j = 3;      
% Set the initial position (initial state): The initial position is 1 (upper left).
    st = 1;
    st1 = 1;
% Generating an array to store the change in Q value in state1
    a = zeros(1,4);

%% Initial configuration of the NN used in DQN (three layers: input layer, hidden layer, and output layer)
% Number of nodes in the hidden layer
    NumHiddenLayer = 5;  
% Setting the activation function of the hidden layer
    HiddenFunc = "sigmoid"; % o͊֐ Sigmoid
% Setting the output function
    OutFunc = "Linear"; % o͊֐ Linear
% Setting Hyperparameters :learning rateC:discount rate
    alpha = 0.001;
    ganma = 0.9;
% Generate hidden layer and output layer tables
    w12 = rand(maze_i * maze_j,NumHiddenLayer);
    b2 = -1*ones(NumHiddenLayer,1);
    w23 = rand(NumHiddenLayer,4);
    b3 = -1*ones(4,1);
% Display Q-table and routes before learning
    disp("Q-table and path before learning");
    print_weight(w12, b2, w23, b3, HiddenFunc, OutFunc, maze_i, maze_j);
    print_route(w12, b2, w23, b3, HiddenFunc, OutFunc, maze_i, maze_j);

%% The following is a study by DQN.
epsilon_tmp = MaxEpisode;
tic;                            % (In the case of time measurement, the line to be commented shall be commented.
for Episode = 1 : MaxEpisode    % Train for the "MaxEpisode" generation.
    % Initialize to the state at the start of the maze (St = 1) for each episode.
        st1 = 1;
    % Definition of epsilon
        epsilon_tmp = epsilon_tmp - 1;  % epsilon-greedy method
        epsilon = epsilon_tmp;
    
    fprintf('Episode = %d, Epsilon = %d\n', Episode, epsilon);

    for t=1:Max_t % The maximum number of actions per generation is "Max_t" times.
        randdata_tmp = uint16(rand*MaxEpisode);
        randdata = randdata_tmp;

        st = st1;                   % Set the current state (the next state in the previous state)
        current_state_num = st;     

        
        % Calculate values (like Q-values) for behavioral guidelines in NN, and output the behavior in the current state and the state of the next step.
            [next_state_num, action, Qt] = sw_move_state(w12, b2, w23, b3, ...
                                            HiddenFunc, OutFunc, epsilon, randdata, ...
                                            current_state_num, maze_i, maze_j);     

        % Determine actions and next states
            at = action;                        
            st1 = next_state_num;               

        % Penalty for staying in the same place to avoid going out of frame.
            if st == st1
                penalty = 5;                    
            elseif t == Max_t
                penalty = 5;
            else
                penalty = 0;                    
            end
        
        % Get reward
        % Negative (penalties) for positions with demons.
            reward = reward_table(st1);         
        
        % Updating weights and biases
            [w12, b2, w23, b3] = sw_update_Q(st, st1, w12, b2, w23, b3, ...
                HiddenFunc, OutFunc, at, reward, penalty, alpha, ganma, maze_i, maze_j);           

        fprintf('Q(%d,%d)->', st, at);
        
        previous_state = st;

        if st1 == 9
            fprintf('S9 Goal\n');
            break;
        end
    end
    fprintf('\n');
    
    % Store the change in Q value in state1 in an array
        st2 = zeros(maze_i*maze_j,1);       % Create One hot input from state
        st2(1) = 1;

    X = st2;                             
    
    z2 = w12.' * X + b2; 
    if(HiddenFunc == "sigmoid")
        a2 = 1./(1+exp(-z2));           % Output of the hidden layer a3 (activation function: sigmoid function)
    end

    z3 = w23.' * a2 + b3;               % Weighted sum of the output layers z3
    if(OutFunc == 'Linear')
        a3 = z3;      % Output of output layer a3 (activation function: lenear function)
    end
        
    a4 = reshape(a3,1,4);
        
    if Episode == 1
        a=a4;
    else
        a = [a;a4];
    end 
end
toc;

disp("Q-table and path after learning");
print_weight(w12, b2, w23, b3, HiddenFunc, OutFunc ,maze_i, maze_j);
print_route(w12, b2, w23, b3, HiddenFunc, OutFunc, maze_i, maze_j);
plot(a); % Graph of change in Q-value of state1
